import math
from functools import partial
from typing import Callable, Optional, Tuple, Type, Union

import folium
import geopandas as gpd
import matplotlib
import matplotlib as mpl
import numpy as np
import pandas as pd
import seaborn as sns
from geojson import LineString
from geopandas import GeoDataFrame
from matplotlib import pyplot as plt
from pandas import DataFrame

import dp_mobility_report.constants as const
from dp_mobility_report.report.html.html_utils import get_centroids

sns.set_theme()


def format(value: Union[float, int], type: Type, ndigits: int = 2) -> Union[float, int]:
    value = type(value)
    if type == float:
        value = round(value, ndigits)
    return value


def histogram(
    hist: Tuple,
    x_axis_label: str,
    hist_alternative: Optional[Tuple] = None,
    min_value: Union[int, float] = None,
    y_axis_label: str = "Frequency",
    margin_of_error: Union[float, list] = None,
    margin_of_error_alternative: Union[float, list] = None,
    rotate_label: bool = False,
    x_axis_type: Type = float,
    ndigits_x_label: int = 2,
) -> mpl.figure.Figure:
    bins = hist[1]
    counts = hist[0]
    counts_alternative = hist_alternative[0] if hist_alternative else None

    # single integers (instead of bin ranges) as x axis labels
    if len(bins) == len(counts):
        labels = np.array([format(bin, x_axis_type) for bin in bins[:-1]])
        if bins[-1] == np.Inf:
            labels = np.append(labels, f"> {labels[-1]}")
        else:
            labels = np.append(labels, format(bins[-1], x_axis_type))
        # needed for margin of error (see below)
        upper_limits = bins

    # bin ranges as labels
    else:
        lower_limits = bins[:-1]
        upper_limits = bins[1:]

        labels = np.array(
            [
                f"[{format(x1, x_axis_type, ndigits_x_label)}\n - \n{format(x2, x_axis_type, ndigits_x_label)})"
                if x2 != np.Inf
                else f"≥ {format(x1, x_axis_type)}"
                for x1, x2 in zip(lower_limits, upper_limits)
            ]
        )

        if upper_limits[-1] != np.Inf:
            labels[-1] = (
                labels[-1][:-1] + "]"
            )  # fix label string of last bin to include last value

    # margin of error only for bins above min value
    if min_value:
        margin_of_error = [
            0 if (min_value > upper_limits[i]) else margin_of_error
            for i in range(0, len(upper_limits))
        ]

    return barchart(
        x=labels,
        y=counts,
        x_axis_label=x_axis_label,
        y_axis_label=y_axis_label,
        y_alternative=counts_alternative,
        margin_of_error=margin_of_error,
        margin_of_error_alternative=margin_of_error_alternative,
        rotate_label=rotate_label,
    )


def barchart(
    x: np.ndarray,
    y: np.ndarray,
    x_axis_label: str,
    y_axis_label: str,
    y_alternative: Optional[np.ndarray] = None,
    margin_of_error: Optional[Union[float, list]] = None,
    margin_of_error_alternative: Optional[Union[float, list]] = None,
    rotate_label: bool = False,
) -> mpl.figure.Figure:
    fig, ax = plt.subplots()
    bar_base = ax.bar(
        x, y, yerr=margin_of_error, align="center", alpha=0.5, capsize=10, label="base"
    )
    if y_alternative is not None:
        bar_alt = ax.bar(
            x,
            y_alternative,
            yerr=margin_of_error_alternative,
            width=0.3,
            align="center",
            alpha=0.5,
            capsize=10,
            color=const.LIGHT_ORANGE,
            ecolor=const.LIGHT_ORANGE,
            label="alternative",
        )
        ax.legend(handles=[bar_base, bar_alt])
    ax.set_ylabel(y_axis_label)
    ax.set_xlabel(x_axis_label)
    plt.xticks(x)
    plt.ylim(bottom=0)

    if rotate_label:
        plt.xticks(rotation=90)
    return fig


def linechart(
    data: DataFrame,
    x: str,
    y: str,
    x_axis_label: str,
    y_axis_label: str,
    x_alternative: Optional[str] = None,
    margin_of_error: float = None,
    add_diagonal: bool = False,
    rotate_label: bool = False,
) -> mpl.figure.Figure:
    fig, ax = plt.subplots()
    if margin_of_error is not None:
        ax.fill_between(
            data[x],
            (data[y] - margin_of_error),
            (data[y] + margin_of_error),
            color="blue",
            alpha=0.1,
        )
    ax.plot(data[x], data[y])
    if x_alternative is not None:
        ax.plot(data[x_alternative], data[y], color=const.LIGHT_ORANGE)
        ax.legend({"base", "alt"})
    ax.set_ylabel(y_axis_label)
    ax.set_xlabel(x_axis_label)
    ax.set_ylim(bottom=0)
    if add_diagonal:
        ax.plot([0, data[x].max()], [0, data[y].max()], const.GREY)

    if rotate_label:
        plt.xticks(rotation=90)

    return fig


def linechart_new(
    data: DataFrame,
    x: str,
    y: str,
    x_axis_label: str,
    y_axis_label: str,
    data_alternative: Optional[DataFrame] = None,
    margin_of_error: float = None,
    margin_of_error_alternative: float = None,
    add_diagonal: bool = False,
    rotate_label: bool = False,
) -> mpl.figure.Figure:
    fig, ax = plt.subplots()
    if margin_of_error is not None:
        ax.fill_between(
            data[x],
            (data[y] - margin_of_error),
            (data[y] + margin_of_error),
            color=const.LIGHT_BLUE,
            alpha=0.1,
        )
    (line_base,) = ax.plot(data[x], data[y], color=const.LIGHT_BLUE, label="base")
    if data_alternative is not None:
        (line_alt,) = ax.plot(
            data_alternative[x],
            data_alternative[y],
            color=const.LIGHT_ORANGE,
            label="alternative",
        )
        ax.legend(handles=[line_base, line_alt])
        if margin_of_error_alternative is not None:
            ax.fill_between(
                data_alternative[x],
                (data_alternative[y] - margin_of_error_alternative),
                (data_alternative[y] + margin_of_error_alternative),
                color=const.LIGHT_ORANGE,
                alpha=0.1,
            )
    ax.set_ylabel(y_axis_label)
    ax.set_xlabel(x_axis_label)
    ax.set_ylim(bottom=0)
    if add_diagonal:
        ax.plot([0, data[x].max()], [0, data[y].max()], const.GREY, alpha=0.4)

    if rotate_label:
        plt.xticks(rotation=90)

    return fig


def multi_linechart(
    data: DataFrame,
    x: str,
    y: str,
    color: str,
    x_axis_label: str,
    y_axis_label: str,
    style: Optional[str] = None,
    hue_order: Optional[list] = None,
    margin_of_error: Optional[float] = None,
) -> mpl.figure.Figure:
    fig = plt.figure(figsize=(9, 6))
    plot = fig.add_subplot(111)
    palette = ["#99065a", "#e289ba", "#2c6a19", "#99cd60"]

    sns.lineplot(
        data=data,
        x=x,
        y=y,
        hue=color,
        ax=plot,
        style=style,
        hue_order=hue_order,
        palette=palette,
    )
    plot.set_ylim(bottom=0)
    plot.set_ylabel(y_axis_label)
    plot.set_xlabel(x_axis_label)

    if margin_of_error is not None:
        for i in range(0, len(hue_order)):
            hue_data = data[data[color] == hue_order[i]]
            plot.fill_between(
                hue_data[x],
                (hue_data[y] - margin_of_error),
                (hue_data[y] + margin_of_error),
                color=palette[i],
                alpha=0.1,
            )
    return fig


def choropleth_map(
    counts_per_tile_gdf: GeoDataFrame,
    fill_color_name: str,
    scale_title: str = "Visit count",
    map: Optional[folium.Map] = None,
    is_cmap_diverging: bool = False,
    cmap: Optional[str] = None,
    min_scale: Optional[Union[int, float]] = None,
    max_scale: Optional[Union[int, float]] = None,
    aliases: Optional[list] = None,
    layer_name: str = "Visits",
    show: bool = True,
) -> folium.Map:
    poly_json = counts_per_tile_gdf.to_json()

    if not map:
        center_x, center_y = _get_center(counts_per_tile_gdf)
        map = _basemap(center_x, center_y)

    min_scale = (
        min_scale
        if min_scale is not None
        else counts_per_tile_gdf[fill_color_name].min()
    )

    max_scale = (
        max_scale
        if max_scale is not None
        else counts_per_tile_gdf[fill_color_name].max()
    )

    # color
    if not cmap:
        cmap = const.DIVERGING_CMAP if is_cmap_diverging else const.STANDARD_CMAP
    mpl_cmap = mpl.colormaps[cmap]

    if is_cmap_diverging:
        norm = mpl.colors.TwoSlopeNorm(vmin=min_scale, vcenter=0, vmax=max_scale)
    else:
        norm = mpl.colors.Normalize(vmin=min_scale, vmax=max_scale)

    def _hex_color(x: Union[float, int], mpl_cmap: mpl.colors.Colormap) -> str:
        rgb = norm(x)
        return matplotlib.colors.rgb2hex(mpl_cmap(rgb))

    def _get_color(x: dict, fill_color_name: str, mpl_cmap: mpl.colors.Colormap) -> str:
        if x["properties"][fill_color_name] is None:
            return "#8c8c8c"
        return _hex_color(x["properties"][fill_color_name], mpl_cmap)

    def _style_function(
        x: dict, fill_color_name: str, mpl_cmap: mpl.colors.Colormap
    ) -> dict:
        return {
            "fillColor": _get_color(x, fill_color_name, mpl_cmap),
            "color": const.GREY,
            "weight": 1.5,
            "fillOpacity": 0.6,
        }

    _style_function_partial = partial(
        _style_function, fill_color_name=fill_color_name, mpl_cmap=mpl_cmap
    )

    if "tile_name" in counts_per_tile_gdf:
        fields = ["tile_id", "tile_name", fill_color_name]
    else:
        fields = ["tile_id", fill_color_name]
    folium.GeoJson(
        poly_json,
        name=layer_name,
        overlay=True,
        show=show,
        style_function=_style_function_partial,
        popup=folium.GeoJsonPopup(fields=fields, aliases=aliases),
    ).add_to(map)

    # colorbar object to create custom legend
    colorbar, ax = plt.subplots(figsize=(6, 1))
    colorbar.subplots_adjust(bottom=0.5)
    ax.grid(False)
    colorbar.colorbar(
        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
        cax=ax,
        orientation="horizontal",
        label=scale_title,
    )
    # TODO: add legend directly to map

    return map, colorbar


def multi_choropleth_map(
    counts_per_tile_timewindow: DataFrame,
    tessellation: GeoDataFrame,
    is_cmap_diverging: bool = False,
    min_scale: Optional[Union[int, float]] = None,
    max_scale: Optional[Union[int, float]] = None,
    vcenter: Optional[Union[int, float]] = 0,
) -> mpl.figure.Figure:
    counts_per_tile_timewindow = tessellation[["tile_id", "geometry"]].merge(
        counts_per_tile_timewindow, left_on="tile_id", right_index=True, how="left"
    )

    # col1: tile_id, col2: geometry
    plot_count = counts_per_tile_timewindow.shape[1] - 2
    plots_per_row = 3
    row_count = math.ceil(plot_count / plots_per_row)
    fig, axes = plt.subplots(row_count, plots_per_row, figsize=(18, 12))

    # upper and lower bound
    if min_scale is None:
        min_scale = counts_per_tile_timewindow.iloc[:, 2:].min().min()
        min_scale = min_scale if not math.isnan(min_scale) else 0
    if max_scale is None:
        max_scale = counts_per_tile_timewindow.iloc[:, 2:].max().max()
        max_scale = max_scale if not math.isnan(max_scale) else 2

    # color
    if is_cmap_diverging:
        cmap = const.DIVERGING_CMAP
        norm = mpl.colors.TwoSlopeNorm(vmin=min_scale, vcenter=vcenter, vmax=max_scale)
    else:
        cmap = const.BASE_CMAP  # STANDARD_CMAP
        norm = mpl.colors.Normalize(vmin=min_scale, vmax=max_scale)

    for i in range(0, plots_per_row * row_count):
        facet_row = math.ceil((i - plots_per_row + 1) / plots_per_row)
        if row_count == 1:
            ax = axes[i % plots_per_row]
        else:
            ax = axes[facet_row][i % plots_per_row]
        ax.axis("off")

        if i < (
            counts_per_tile_timewindow.shape[1] - 2
        ):  # there might be more subplots than data - skip in that case
            column_name = counts_per_tile_timewindow.columns[i + 2]
            counts_per_tile_timewindow.iloc[:, [i + 2, 1]].plot(
                column=column_name,
                cmap=cmap,
                norm=norm,
                linewidth=0.1,
                ax=ax,
                edgecolor="#FFFFFF",
                missing_kwds={
                    "color": const.LIGHT_GREY,
                },
            )
            ax.set_title(column_name)

    # Create colorbar as a legend
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm._A = []  # add the colorbar to the figure
    # set the range for the choropleth
    plt.rcParams["axes.grid"] = False  # silence matplotlib warning
    fig.colorbar(sm, ax=axes)
    plt.rcParams["axes.grid"] = True

    return fig


def _get_center(gdf: GeoDataFrame) -> Tuple:
    center_x = (gdf.total_bounds[0] + gdf.total_bounds[2]) / 2
    center_y = (gdf.total_bounds[1] + gdf.total_bounds[3]) / 2
    return (center_x, center_y)


def _basemap(center_x: float, center_y: float, zoom_start: int = 10) -> folium.Map:
    return folium.Map(
        tiles="CartoDB positron",
        attr='&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>, &copy; <a href="https://carto.com/attributions">CARTO</a>',
        zoom_start=zoom_start,
        location=[center_y, center_x],
        control_scale=True,
    )


def ranking(
    x: np.ndarray,
    x_axis_label: str,
    x_alternative: Optional[np.ndarray] = None,
    y_labels: list = None,
    margin_of_error: Optional[float] = None,
    margin_of_error_alternative: Optional[float] = None,
) -> mpl.figure.Figure:
    y = list(range(1, len(x) + 1))[::-1]
    y_labels = y if y_labels is None else y_labels
    fig, ax = plt.subplots()
    bar_base = ax.errorbar(
        x,
        y,
        xerr=margin_of_error,
        fmt="o",
        ecolor="lightblue",
        elinewidth=5,
        capsize=10,
        label="base",
    )
    if x_alternative is not None:
        bar_alt = ax.errorbar(
            x_alternative,
            y,
            xerr=margin_of_error_alternative,
            fmt="o",
            ecolor=const.LIGHT_ORANGE,
            elinewidth=5,
            capsize=10,
            label="alternative",
        )
        ax.legend(handles=[bar_base, bar_alt])
    ax.set_yticks(y)
    ax.set_yticklabels(y_labels)
    ax.set_ylabel("Rank")
    ax.set_xlabel(x_axis_label)
    return fig


def flows(
    basemap: folium.Map,
    data: pd.DataFrame,
    tessellation: gpd.GeoDataFrame,
    flow_color: str = const.DARK_BLUE,
    marker_color: str = const.LIGHT_BLUE,
    layer_name: str = "flows",
) -> folium.Map:
    centroids = get_centroids(tessellation)
    mean_flows = data["flow"].mean()

    def _flow_style_function(weight: float, color: str) -> Callable:
        return lambda feature: {
            "color": color,
            "weight": 5 * weight**0.5,
            "opacity": 0.65,
        }

    feature_group = folium.FeatureGroup(name=layer_name)

    origin_groups = data.groupby(by=const.ORIGIN)
    for origin, OD in origin_groups:
        lonO, latO = centroids[origin]

        for destination, flow in OD[[const.DESTINATION, const.FLOW]].values:
            lonD, latD = centroids[destination]
            gjc = LineString([(lonO, latO), (lonD, latD)])

            fgeojson = folium.GeoJson(
                gjc,
                style_function=_flow_style_function(
                    weight=flow / mean_flows, color=flow_color
                ),
            )

            popup = folium.Popup(
                f"flow from {origin} to {destination}: {int(flow)}", max_width=300
            )
            fgeojson = fgeojson.add_child(popup)
            feature_group.add_child(fgeojson)

        # Plot marker
        fmarker = folium.CircleMarker(
            [latO, lonO],
            radius=5,
            weight=2,
            color=marker_color,
            fill=True,
            fill_color=marker_color,
        )
        T_D = [
            [destination, int(flow)]
            for destination, flow in OD[[const.DESTINATION, const.FLOW]].values
        ]
        trips_info = "<br/>".join(
            [
                f"flow to {destination}: {flow}"
                for destination, flow in sorted(T_D, reverse=True)[:5]
            ]
        )
        name = f"origin: {origin}"
        popup = folium.Popup(name + "<br/>" + trips_info, max_width=300)
        fmarker = fmarker.add_child(popup)
        feature_group.add_child(fmarker)

    feature_group.add_to(basemap)

    return basemap



import math
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import geopandas as gpd
import jinja2
import numpy as np
from pandas import Series

from dp_mobility_report import constants as const

if TYPE_CHECKING:
    from dp_mobility_report import BenchmarkReport

# Initializing Jinja
package_loader = jinja2.PackageLoader(
    "dp_mobility_report", "report/html/html_templates"
)
jinja2_env = jinja2.Environment(
    lstrip_blocks=True, trim_blocks=True, loader=package_loader
)


def get_template(template_name: str) -> jinja2.Template:
    return jinja2_env.get_template(template_name)


def render_summary(summary: Series) -> str:
    summary_list = [
        {"name": "Min.", "value": fmt(summary["min"])},
        {"name": "Max.", "value": fmt(summary["max"])},
    ]
    if "25%" in summary:
        summary_list.insert(1, {"name": "75%", "value": fmt(summary["75%"])})
        summary_list.insert(1, {"name": "Median", "value": fmt(summary["50%"])})
        summary_list.insert(1, {"name": "25%", "value": fmt(summary["25%"])})

    if "mean" in summary:
        summary_list.insert(0, {"name": "Mean", "value": fmt(summary["mean"])})

    template_table = jinja2_env.get_template("table.html")
    summary_html = template_table.render(rows=summary_list)
    return summary_html


def render_benchmark_summary(
    summary_base: Series,
    summary_alternative: Series,
    target_type: Optional[type] = None,
) -> str:
    summary_list = [
        {
            "name": "Min.",
            "value": (
                fmt(summary_base["min"], target_type),
                fmt(summary_alternative["min"], target_type),
            ),
        },
        {
            "name": "Max.",
            "value": (
                fmt(summary_base["max"], target_type),
                fmt(summary_alternative["max"], target_type),
            ),
        },
    ]
    if "25%" in summary_base:
        summary_list.insert(
            1,
            {
                "name": "75%",
                "value": (
                    fmt(summary_base["75%"], target_type),
                    fmt(summary_alternative["75%"], target_type),
                ),
            },
        )
        summary_list.insert(
            1,
            {
                "name": "Median",
                "value": (
                    fmt(summary_base["50%"], target_type),
                    fmt(summary_alternative["50%"], target_type),
                ),
            },
        )
        summary_list.insert(
            1,
            {
                "name": "25%",
                "value": (
                    fmt(summary_base["25%"], target_type),
                    fmt(summary_alternative["25%"], target_type),
                ),
            },
        )

    if "mean" in summary_base:
        summary_list.insert(
            0,
            {
                "name": "Mean",
                "value": (
                    fmt(summary_base["mean"], target_type),
                    fmt(summary_alternative["mean"], target_type),
                ),
            },
        )

    template_table = jinja2_env.get_template("table_benchmark.html")
    summary_html = template_table.render(rows=summary_list)
    return summary_html


def render_user_input_info(
    max_value: Optional[Union[int, float]], bin_size: Optional[Union[int, float]]
) -> str:
    return f"""User configuration for histogram chart: <br>
            maximum value: {max_value} <br>
            bin size: {bin_size}"""


def render_moe_info(margin_of_error: int) -> str:
    return """To provide privacy, quartile values are not necessarily the true values but, e.g., instead of the true maximum value the 
        second or third highest value is displayed.
        This is achieved by the so-called exponential mechanism, where a value is drawn based on probabilites defined by the privacy budget. 
        Generally, a value closer to the true value has a higher chance of being drawn."""
    # The true quartile values lie with a <b>95% chance within ± {margin_of_error} records</b> away from the true values.""" # TODO: margin_of_error reveals true record count?


def fmt(value: Any, target_type: Optional[type] = None) -> Any:
    if target_type and (value is not None):
        value = target_type(value)
    if isinstance(value, (float, np.floating)):
        if math.isinf(value) or np.isnan(value):
            return "not defined"
        value = round(value, 2)
    if isinstance(value, (float, np.floating, int, np.integer)) and not isinstance(
        value, bool
    ):
        value = f"{value:,}"
    return value


def fmt_moe(margin_of_error: Optional[float]) -> float:
    if (margin_of_error is None) or (margin_of_error == 0):
        return 0
    return round(margin_of_error, 1)


def fmt_config(value: Union[dict, list]) -> str:
    def _join_tuple_string(strings_tuple: tuple) -> str:
        return ": ".join(strings_tuple)

    analysis_format = {
        const.DS_STATISTICS: "Dataset statistics",
        const.MISSING_VALUES: "Missing values",
        const.TRIPS_OVER_TIME: "Trips over time",
        const.TRIPS_PER_WEEKDAY: "Trips per weekday",
        const.TRIPS_PER_HOUR: "Trips per hour",
        const.VISITS_PER_TILE: "Visits per tile",
        const.VISITS_PER_TIME_TILE: "Visits per time tile",
        const.VISITS_PER_TILE_OUTLIERS: "Visits per tile outliers",
        const.VISITS_PER_TILE_RANKING: "Visits per tile ranking",
        const.VISITS_PER_TILE_QUARTILES: "Visits per tile quartiles",
        const.OD_FLOWS: "OD flows",
        const.OD_FLOWS_RANKING: "OD flows ranking",
        const.OD_FLOWS_QUARTILES: "Od flows quartiles",
        const.TRAVEL_TIME: "Travel time",
        const.TRAVEL_TIME_QUARTILES: "Travel time quartiles",
        const.JUMP_LENGTH: "Jump length",
        const.JUMP_LENGTH_QUARTILES: "Jump length quartiles",
        const.TRIPS_PER_USER: "Trips per user",
        const.TRIPS_PER_USER_QUARTILES: "Trips per user quartiles",
        const.USER_TIME_DELTA: "User time delta",
        const.USER_TIME_DELTA_QUARTILES: "User time delta quartiles",
        const.RADIUS_OF_GYRATION: "Radius of gyration",
        const.RADIUS_OF_GYRATION_QUARTILES: "Radius of gyration quartiles",
        const.USER_TILE_COUNT: "User tile count",
        const.USER_TILE_COUNT_QUARTILES: "User tile count quartiles",
        const.MOBILITY_ENTROPY: "Mobility entropy",
        const.MOBILITY_ENTROPY_QUARTILES: "Mobility entropy quartiles",
    }

    if value == {}:
        return "Evenly distributed"
    if value == []:
        return "None"
    if isinstance(value, dict):
        zipped = zip(
            [analysis_format[s] for s in value.keys()], [str(v) for v in value.values()]
        )
        return ", ".join(list(map(_join_tuple_string, zipped)))
    if isinstance(value, list):
        return ", ".join([analysis_format[s] for s in value])


def render_eps(value: Optional[float]) -> Optional[float]:
    if value is None:
        return None
    else:
        return round(value, 4)


def get_centroids(tessellation: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    centroids = (
        tessellation.geometry.to_crs(3857).centroid.to_crs(4326).apply(lambda x: x.xy)
    )
    lngs = [c[0].pop() for c in centroids]
    lats = [c[1].pop() for c in centroids]
    return dict(zip(tessellation[const.TILE_ID], zip(lngs, lats)))


def all_available_measures(
    analysis_name: str, benchmarkreport: "BenchmarkReport"
) -> dict:
    available_measures: Dict[str, Union[str, List[str]]] = {}

    if analysis_name in benchmarkreport.smape:
        available_measures[const.SMAPE] = str(fmt(benchmarkreport.smape[analysis_name]))
    if analysis_name in benchmarkreport.jsd:
        available_measures[const.JSD] = str(fmt(benchmarkreport.jsd[analysis_name]))
    if analysis_name in benchmarkreport.kld:
        available_measures[const.KLD] = str(fmt(benchmarkreport.kld[analysis_name]))
    if analysis_name in benchmarkreport.emd:
        available_measures[const.EMD] = str(fmt(benchmarkreport.emd[analysis_name]))
    if analysis_name in benchmarkreport.kt:
        available_measures[const.KT] = [
            f"Top {topn_i}: {fmt(kt_i)}"
            for kt_i, topn_i in zip(
                benchmarkreport.kt[analysis_name], benchmarkreport.top_n_ranking
            )
        ]
    if analysis_name in benchmarkreport.top_n_cov:
        available_measures[const.TOP_N_COV] = [
            f"Top {topn_i}: {fmt(topcov_i * 100)}%"
            for topcov_i, topn_i in zip(
                benchmarkreport.top_n_cov[analysis_name], benchmarkreport.top_n_ranking
            )
        ]

    return available_measures



from typing import TYPE_CHECKING

from pandas import DataFrame
from tqdm.auto import tqdm

if TYPE_CHECKING:
    from dp_mobility_report import DpMobilityReport

from dp_mobility_report import constants as const
from dp_mobility_report.model import (
    od_analysis,
    overview,
    place_analysis,
    user_analysis,
)


def report_elements(dpmreport: "DpMobilityReport") -> dict:

    report: dict = {}

    # budget sum consists of number of elements weighted with 1 (no special budget assigned)
    # plus budget for each custom assigned analysis budget
    budget_split_sum = (
        len(const.ELEMENTS)
        - len(dpmreport.analysis_exclusion)
        - len(dpmreport.budget_split.keys())
        + sum(dpmreport.budget_split.values())
    )

    # get privacy budget for each report element
    if dpmreport.privacy_budget is None or dpmreport.evalu:
        eps_factor = dpmreport.privacy_budget
    else:
        eps_factor = (
            dpmreport.privacy_budget / budget_split_sum
            if budget_split_sum > 0
            else dpmreport.privacy_budget
        )

    with tqdm(  # progress bar
        total=4, desc="Create report", disable=dpmreport.disable_progress_bar
    ) as pbar:

        report = {**report, **add_overview_elements(dpmreport, eps_factor)}
        pbar.update()

        report = {
            **report,
            **add_place_analysis_elements(dpmreport, eps_factor),
        }
        pbar.update()

        if not set(const.OD_ELEMENTS).issubset(dpmreport.analysis_exclusion):
            _od_shape = od_analysis.get_od_shape(dpmreport.df)
            report = {
                **report,
                **add_od_analysis_elements(dpmreport, _od_shape, eps_factor),
            }
        pbar.update()

        report = {**report, **add_user_analysis_elements(dpmreport, eps_factor)}
        pbar.update()

    return report


def _get_eps(eps_factor: float, analysis_name: str, budget_split: dict) -> float:
    if eps_factor is None:
        return None
    elif analysis_name not in budget_split:
        # analysis gets default factor of 1
        return eps_factor
    else:
        # else epsilon is multiplied by the respective configured factor
        return budget_split[analysis_name] * eps_factor


def add_overview_elements(dpmreport: "DpMobilityReport", eps_factor: float) -> dict:
    overview_elements: dict = {}

    if (const.DS_STATISTICS in const.OVERVIEW_ELEMENTS) and (
        const.DS_STATISTICS not in dpmreport.analysis_exclusion
    ):
        overview_elements[const.DS_STATISTICS] = overview.get_dataset_statistics(
            dpmreport, _get_eps(eps_factor, const.DS_STATISTICS, dpmreport.budget_split)
        )

    if (const.MISSING_VALUES in const.OVERVIEW_ELEMENTS) and (
        const.MISSING_VALUES not in dpmreport.analysis_exclusion
    ):
        overview_elements[const.MISSING_VALUES] = overview.get_missing_values(
            dpmreport,
            _get_eps(eps_factor, const.MISSING_VALUES, dpmreport.budget_split),
        )

    if (const.TRIPS_OVER_TIME in const.OVERVIEW_ELEMENTS) and (
        const.TRIPS_OVER_TIME not in dpmreport.analysis_exclusion
    ):
        overview_elements[const.TRIPS_OVER_TIME] = overview.get_trips_over_time(
            dpmreport,
            _get_eps(eps_factor, const.TRIPS_OVER_TIME, dpmreport.budget_split),
        )

    if (const.TRIPS_PER_WEEKDAY in const.TRIPS_PER_WEEKDAY) and (
        const.TRIPS_PER_WEEKDAY not in dpmreport.analysis_exclusion
    ):
        overview_elements[const.TRIPS_PER_WEEKDAY] = overview.get_trips_per_weekday(
            dpmreport,
            _get_eps(eps_factor, const.TRIPS_PER_WEEKDAY, dpmreport.budget_split),
        )

    if (const.TRIPS_PER_HOUR in const.OVERVIEW_ELEMENTS) and (
        const.TRIPS_PER_HOUR not in dpmreport.analysis_exclusion
    ):
        overview_elements[const.TRIPS_PER_HOUR] = overview.get_trips_per_hour(
            dpmreport,
            _get_eps(eps_factor, const.TRIPS_PER_HOUR, dpmreport.budget_split),
        )
    return overview_elements


def add_place_analysis_elements(
    dpmreport: "DpMobilityReport",
    eps_factor: float,
) -> dict:
    place_analysis_elements: dict = {}

    if (const.VISITS_PER_TILE in const.PLACE_ELEMENTS) and (
        const.VISITS_PER_TILE not in dpmreport.analysis_exclusion
    ):
        place_analysis_elements[
            const.VISITS_PER_TILE
        ] = place_analysis.get_visits_per_tile(
            dpmreport,
            _get_eps(eps_factor, const.VISITS_PER_TILE, dpmreport.budget_split),
        )

    if (const.VISITS_PER_TIME_TILE in const.PLACE_ELEMENTS) and (
        const.VISITS_PER_TIME_TILE not in dpmreport.analysis_exclusion
    ):
        place_analysis_elements[
            const.VISITS_PER_TIME_TILE
        ] = place_analysis.get_visits_per_time_tile(
            dpmreport,
            _get_eps(eps_factor, const.VISITS_PER_TIME_TILE, dpmreport.budget_split),
        )
    return place_analysis_elements


def add_od_analysis_elements(
    dpmreport: "DpMobilityReport",
    _od_shape: DataFrame,
    eps_factor: float,
) -> dict:
    od_analysis_elements: dict = {}

    if (const.OD_FLOWS in const.OD_ELEMENTS) and (
        const.OD_FLOWS not in dpmreport.analysis_exclusion
    ):
        od_analysis_elements[const.OD_FLOWS] = od_analysis.get_od_flows(
            _od_shape,
            dpmreport,
            _get_eps(eps_factor, const.OD_FLOWS, dpmreport.budget_split),
        )

    if (const.TRAVEL_TIME in const.OD_ELEMENTS) and (
        const.TRAVEL_TIME not in dpmreport.analysis_exclusion
    ):
        od_analysis_elements[const.TRAVEL_TIME] = od_analysis.get_travel_time(
            _od_shape,
            dpmreport,
            _get_eps(eps_factor, const.TRAVEL_TIME, dpmreport.budget_split),
        )

    if (const.JUMP_LENGTH in const.OD_ELEMENTS) and (
        const.JUMP_LENGTH not in dpmreport.analysis_exclusion
    ):
        od_analysis_elements[const.JUMP_LENGTH] = od_analysis.get_jump_length(
            _od_shape,
            dpmreport,
            _get_eps(eps_factor, const.JUMP_LENGTH, dpmreport.budget_split),
        )

    return od_analysis_elements


def add_user_analysis_elements(
    dpmreport: "DpMobilityReport", eps_factor: float
) -> dict:
    user_analysis_elements = {}
    if (const.TRIPS_PER_USER in const.USER_ELEMENTS) and (
        const.TRIPS_PER_USER not in dpmreport.analysis_exclusion
    ):
        user_analysis_elements[const.TRIPS_PER_USER] = user_analysis.get_trips_per_user(
            dpmreport,
            _get_eps(eps_factor, const.TRIPS_PER_USER, dpmreport.budget_split),
        )

    if (const.USER_TIME_DELTA in const.USER_ELEMENTS) and (
        const.USER_TIME_DELTA not in dpmreport.analysis_exclusion
    ):
        user_analysis_elements[
            const.USER_TIME_DELTA
        ] = user_analysis.get_user_time_delta(
            dpmreport,
            _get_eps(eps_factor, const.USER_TIME_DELTA, dpmreport.budget_split),
        )

    if (const.RADIUS_OF_GYRATION in const.USER_ELEMENTS) and (
        const.RADIUS_OF_GYRATION not in dpmreport.analysis_exclusion
    ):
        user_analysis_elements[
            const.RADIUS_OF_GYRATION
        ] = user_analysis.get_radius_of_gyration(
            dpmreport,
            _get_eps(eps_factor, const.RADIUS_OF_GYRATION, dpmreport.budget_split),
        )

    if (const.USER_TILE_COUNT in const.USER_ELEMENTS) and (
        const.USER_TILE_COUNT not in dpmreport.analysis_exclusion
    ):
        user_analysis_elements[
            const.USER_TILE_COUNT
        ] = user_analysis.get_user_tile_count(
            dpmreport,
            _get_eps(eps_factor, const.USER_TILE_COUNT, dpmreport.budget_split),
        )

    if (const.MOBILITY_ENTROPY in const.USER_ELEMENTS) and (
        const.MOBILITY_ENTROPY not in dpmreport.analysis_exclusion
    ):
        user_analysis_elements[
            const.MOBILITY_ENTROPY
        ] = user_analysis.get_mobility_entropy(
            dpmreport,
            _get_eps(eps_factor, const.MOBILITY_ENTROPY, dpmreport.budget_split),
        )
    return user_analysis_elements



"""Top-level package for Mobility Report dp_mobility_report."""

from dp_mobility_report.benchmark.b_utils import default_measure_selection  # noqa: F401
from dp_mobility_report.benchmark.benchmarkreport import BenchmarkReport  # noqa: F401
from dp_mobility_report.dpmreport import DpMobilityReport  # noqa: F401

__version__ = "0.2.2"



import warnings
from typing import List, Tuple

import numpy as np
import pandas as pd

from dp_mobility_report import constants as const
from dp_mobility_report.benchmark.b_utils import default_measure_selection
from dp_mobility_report.model.section import DfSection


# only select analyses that are present in both reports
def combine_analysis_exclusion(
    analysis_exclusion_alternative: List[str], analysis_exclusion_base: List[str]
) -> List[str]:
    analysis_exclusion = list(
        set(analysis_exclusion_alternative + analysis_exclusion_base)
    )
    analysis_exclusion.sort(key=lambda i: const.ELEMENTS.index(i))
    return analysis_exclusion


def _resample_to_prec(
    df: pd.DataFrame, current_precision: str, target_precision: str
) -> pd.DataFrame:
    if current_precision == target_precision:
        return df
    precision_names = {
        const.PREC_MONTH: "M",
        const.PREC_WEEK: "W-MON",
        const.PREC_DATE: "1D",
    }
    df[const.DATETIME] = pd.to_datetime(df[const.DATETIME])
    df_resampled = (
        df.set_index(const.DATETIME)
        .resample(precision_names[target_precision], label="left")
        .sum()
        .reset_index()
    )
    df_resampled[const.DATETIME] = df_resampled[const.DATETIME].dt.date
    return df_resampled


def unify_trips_over_time(base: DfSection, alternative: DfSection) -> None:
    if base.datetime_precision != alternative.datetime_precision:
        if const.PREC_MONTH in [
            base.datetime_precision,
            alternative.datetime_precision,
        ]:
            base.data = _resample_to_prec(
                base.data, base.datetime_precision, const.PREC_MONTH
            )
            base.datetime_precision = const.PREC_MONTH
            alternative.data = _resample_to_prec(
                alternative.data, alternative.datetime_precision, const.PREC_MONTH
            )
            alternative.datetime_precision = const.PREC_MONTH
        elif const.PREC_WEEK in [
            base.datetime_precision,
            alternative.datetime_precision,
        ]:
            base.data = _resample_to_prec(
                base.data, base.datetime_precision, const.PREC_WEEK
            )
            base.datetime_precision = const.PREC_WEEK
            alternative.data = _resample_to_prec(
                alternative.data, alternative.datetime_precision, const.PREC_WEEK
            )
            alternative.datetime_precision = const.PREC_WEEK

    missing_dates_base = list(
        set(alternative.data[const.DATETIME]) - set(base.data[const.DATETIME])
    )
    missing_dates_alternative = list(
        set(base.data[const.DATETIME]) - set(alternative.data[const.DATETIME])
    )
    base.data = pd.concat(
        [
            base.data,
            pd.DataFrame(
                data={const.DATETIME: missing_dates_base, "trip_count": 0, "trips": 0}
            ),
        ]
    ).sort_values(const.DATETIME)
    alternative.data = pd.concat(
        [
            alternative.data,
            pd.DataFrame(
                data={
                    const.DATETIME: missing_dates_alternative,
                    "trip_count": 0,
                    "trips": 0,
                }
            ),
        ]
    ).sort_values(const.DATETIME)


def _pad_missing_values(
    hist_bins: np.array, hist_values: np.array, bins_union: np.array
) -> np.array:
    missing_bins = [i for i in bins_union if i not in hist_bins]
    n_missing_left = sum(missing_bins < min(hist_bins))
    n_missing_right = sum(missing_bins > max(hist_bins))
    return np.append(
        np.zeros(n_missing_left), np.append(hist_values, np.zeros(n_missing_right))
    )


# check if histograms of both reports have similar bins
# if bins are missing they can only be at the lower end of the bins as bin_range and max_bins are similar
def unify_histogram_bins(
    report_base: dict, report_alternative: dict, analysis_exclusion: list
) -> Tuple[dict, dict]:
    histograms = [
        const.TRAVEL_TIME,
        const.JUMP_LENGTH,
        const.RADIUS_OF_GYRATION,
        const.USER_TIME_DELTA,
        const.USER_TILE_COUNT,
        const.TRIPS_OVER_TIME,
        # const.TRIPS_PER_USER, TODO: how to handle?
    ]
    # const.MOBILITY_ENTROPY should already be similar: 0-1 with 0.1 bin range

    for hist in histograms:
        if hist not in analysis_exclusion:
            if hist == const.TRIPS_OVER_TIME:
                unify_trips_over_time(report_base[hist], report_alternative[hist])

            else:
                hist_base_values, hist_base_bins = report_base[hist].data
                hist_alternative_values, hist_alternative_bins = report_alternative[
                    hist
                ].data

                bins_union = np.union1d(hist_alternative_bins, hist_base_bins)

                hist_base_values = _pad_missing_values(
                    hist_base_bins, hist_base_values, bins_union
                )
                hist_alternative_values = _pad_missing_values(
                    hist_alternative_bins, hist_alternative_values, bins_union
                )

                report_base[hist].data = (hist_base_values, bins_union)
                report_alternative[hist].data = (hist_alternative_values, bins_union)
    return report_base, report_alternative


def validate_measure_selection(
    measure_selection: dict, analysis_exclusion: list
) -> dict:

    default_selection = default_measure_selection()

    if not isinstance(measure_selection, dict):
        raise TypeError("'measure selection' is not a dictionary")

    for key in list(measure_selection.keys()):
        if key not in default_selection.keys():
            warnings.warn(
                f"{key} is not a valid key (analysis) and will be removed from the measure selection."
            )
            del measure_selection[key]

    for key, value in list(measure_selection.items()):

        if value not in const.SIMILARITY_MEASURES:
            warnings.warn(
                f"{value} is not a valid value (similarity measure) and will be removed from the measure selection."
            )
            del measure_selection[key]

    measure_selection = {**default_selection, **measure_selection}

    for analysis in analysis_exclusion:
        all_analysis = [
            item for item in measure_selection.keys() if item.startswith(analysis)
        ]
        for item in all_analysis:
            del measure_selection[item]

    return measure_selection


def validate_top_n_ranking(top_n_ranking: List[int]) -> List[int]:
    if not isinstance(top_n_ranking, list):
        raise TypeError(
            f"Input parameter top_n_ranking is not a list. Instead: {top_n_ranking}."
        )
    for x in top_n_ranking:
        if not isinstance(x, int):
            raise TypeError(
                f"Input parameter top_n_ranking doest not consist of ints. Instead: {top_n_ranking}."
            )
        if x <= 0:
            raise TypeError(
                f"Input parameter top_n_ranking doest not consist of ints greater 0. Instead: {top_n_ranking}."
            )
    return top_n_ranking



import os
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

import matplotlib.pyplot as plt
import pandas as pd
from geopandas import GeoDataFrame

if TYPE_CHECKING:
    from dp_mobility_report import BenchmarkReport, DpMobilityReport

import folium

from dp_mobility_report import constants as const
from dp_mobility_report.benchmark.similarity_measures import symmetric_perc_error
from dp_mobility_report.model.section import DfSection
from dp_mobility_report.report.html.html_utils import (
    all_available_measures,
    fmt_moe,
    get_template,
    render_benchmark_summary,
    render_eps,
    render_summary,
)
from dp_mobility_report.visualization import plot, v_utils


def render_place_analysis(
    dpmreport: "DpMobilityReport",
    temp_map_folder: Path,
    output_filename: str,
) -> str:
    THRESHOLD = 0.2  # 20%
    args: dict = {}
    report = dpmreport.report
    tessellation = dpmreport.tessellation

    args[
        "privacy_info"
    ] = f"""Tiles below a certain threshold are grayed out: 
        Due to the applied noise, tiles with a low visit count are likely to contain a high percentage of noise. 
        For usability reasons, such unrealistic values are grayed out. 
        More specifically: The threshold is set so that values for tiles with a 5% chance (or higher) of deviating more than {round(THRESHOLD * 100)} percentage points from the estimated value are not shown."""
    args["output_filename"] = output_filename

    if const.VISITS_PER_TILE not in dpmreport.analysis_exclusion:
        args["visits_per_tile_eps"] = render_eps(
            report[const.VISITS_PER_TILE].privacy_budget
        )
        args["visits_per_tile_moe"] = fmt_moe(
            report[const.VISITS_PER_TILE].margin_of_error_laplace
        )

        args["points_outside_tessellation_info"] = render_points_outside_tess(
            report[const.VISITS_PER_TILE]
        )
        args["visits_per_tile_legend"] = render_visits_per_tile(
            report[const.VISITS_PER_TILE], tessellation, THRESHOLD, temp_map_folder
        )
        quartiles = report[const.VISITS_PER_TILE].quartiles.round()

        args["visits_per_tile_summary_table"] = render_summary(
            quartiles.astype(int),  # extrapolate visits from dp record count
        )
        args["visits_per_tile_cumsum_linechart"] = render_visits_per_tile_cumsum(
            report[const.VISITS_PER_TILE],
            diagonal=True,
        )
        args["most_freq_tiles_ranking"] = render_most_freq_tiles_ranking(
            report[const.VISITS_PER_TILE],
        )

    if const.VISITS_PER_TIME_TILE not in dpmreport.analysis_exclusion:
        args["visits_per_tile_timewindow_eps"] = render_eps(
            report[const.VISITS_PER_TIME_TILE].privacy_budget
        )
        args["visits_per_tile_timewindow_moe"] = fmt_moe(
            report[const.VISITS_PER_TIME_TILE].margin_of_error_laplace
        )

        args["visits_per_tile_time_map"] = render_visits_per_time_tile(
            report[const.VISITS_PER_TIME_TILE], tessellation, THRESHOLD
        )

    template_structure = get_template("place_analysis_segment.html")

    return template_structure.render(args)


def render_benchmark_place_analysis(
    benchmark: "BenchmarkReport",
    temp_map_folder: Path,
    output_filename: str,
) -> str:
    args: dict = {}
    report_base = benchmark.report_base.report
    report_alternative = benchmark.report_alternative.report
    tessellation = benchmark.report_base.tessellation
    template_measures = get_template("similarity_measures.html")

    args["output_filename"] = output_filename

    if const.VISITS_PER_TILE not in benchmark.analysis_exclusion:
        args["visits_per_tile_eps"] = (
            render_eps(report_base[const.VISITS_PER_TILE].privacy_budget),
            render_eps(report_alternative[const.VISITS_PER_TILE].privacy_budget),
        )
        args["visits_per_tile_moe"] = (
            fmt_moe(report_base[const.VISITS_PER_TILE].margin_of_error_laplace),
            fmt_moe(report_alternative[const.VISITS_PER_TILE].margin_of_error_laplace),
        )

        args["points_outside_tessellation_info_base"] = render_points_outside_tess(
            report_base[const.VISITS_PER_TILE]
        )
        args[
            "points_outside_tessellation_info_alternative"
        ] = render_points_outside_tess(report_alternative[const.VISITS_PER_TILE])
        args["visits_per_tile_legend"] = render_benchmark_visits_per_tile(
            report_base[const.VISITS_PER_TILE],
            report_alternative[const.VISITS_PER_TILE],
            tessellation,
            temp_map_folder,
        )
        quartiles_base = report_base[const.VISITS_PER_TILE].quartiles.round()
        quartiles_alternative = report_alternative[
            const.VISITS_PER_TILE
        ].quartiles.round()
        args["visits_per_tile_summary_table"] = render_benchmark_summary(
            quartiles_base,
            quartiles_alternative,
            target_type=int,
        )
        args["visits_per_tile_cumsum_linechart"] = render_visits_per_tile_cumsum(
            report_base[const.VISITS_PER_TILE],
            report_alternative[const.VISITS_PER_TILE],
            diagonal=False,
        )
        args["most_freq_tiles_ranking"] = render_most_freq_tiles_ranking_benchmark(
            report_base[const.VISITS_PER_TILE],
            report_alternative[const.VISITS_PER_TILE],
        )
        args["visits_per_tile_measure"] = template_measures.render(
            all_available_measures(const.VISITS_PER_TILE, benchmark)
        )

        args["visits_per_tile_summary_measure"] = template_measures.render(
            all_available_measures(const.VISITS_PER_TILE_QUARTILES, benchmark)
        )

        args["visits_per_tile_ranking_measure"] = template_measures.render(
            {
                **all_available_measures(const.VISITS_PER_TILE_RANKING, benchmark),
                **{"top_n_object": "locations"},
            }
        )

    if const.VISITS_PER_TIME_TILE not in benchmark.analysis_exclusion:
        args["visits_per_time_tile_eps"] = (
            render_eps(report_base[const.VISITS_PER_TIME_TILE].privacy_budget),
            render_eps(report_alternative[const.VISITS_PER_TIME_TILE].privacy_budget),
        )
        args["visits_per_time_tile_moe"] = (
            fmt_moe(report_base[const.VISITS_PER_TIME_TILE].margin_of_error_laplace),
            fmt_moe(
                report_alternative[const.VISITS_PER_TIME_TILE].margin_of_error_laplace
            ),
        )

        args["visits_per_tile_time_map"] = render_visits_per_time_tile_benchmark(
            report_base[const.VISITS_PER_TIME_TILE],
            report_alternative[const.VISITS_PER_TIME_TILE],
            tessellation,
        )

        args["visits_per_time_tile_measure"] = template_measures.render(
            all_available_measures(const.VISITS_PER_TIME_TILE, benchmark)
        )

    template_structure = get_template("place_analysis_segment_benchmark.html")

    return template_structure.render(args)


def render_points_outside_tess(visits_per_tile: DfSection) -> str:
    return f"""{round(visits_per_tile.n_outliers)} ({round(visits_per_tile.n_outliers / (visits_per_tile.data["visits"].sum()
 + visits_per_tile.n_outliers) * 100, 2)}%) points are outside the given tessellation 
    (95% confidence interval ± {round(visits_per_tile.margin_of_error_laplace)})."""


def render_visits_per_tile(
    visits_per_tile: DfSection,
    tessellation: GeoDataFrame,
    threshold: float,
    temp_map_folder: Path,
) -> str:

    # merge count and tessellation
    counts_per_tile_gdf = pd.merge(
        tessellation,
        visits_per_tile.data[[const.TILE_ID, "visits"]],
        how="left",
        left_on=const.TILE_ID,
        right_on=const.TILE_ID,
    )

    # filter visit counts above error threshold
    moe_deviation = (
        visits_per_tile.margin_of_error_laplace / counts_per_tile_gdf["visits"]
    )

    counts_per_tile_gdf.loc[moe_deviation > threshold, "visits"] = None
    map, legend = plot.choropleth_map(
        counts_per_tile_gdf,
        "visits",
        scale_title="number of visits",
        aliases=["Tile ID", "Tile Name", "number of visits"],
        cmap=const.BASE_CMAP,
    )

    map.save(os.path.join(temp_map_folder, "visits_per_tile_map.html"))

    legend_html = v_utils.fig_to_html(legend)
    plt.close()
    return legend_html


def render_benchmark_visits_per_tile(
    visits_per_tile_base: DfSection,
    visits_per_tile_alternative: DfSection,
    tessellation: GeoDataFrame,
    temp_map_folder: Path,
) -> List[str]:

    visits_per_tile_base_sorted = visits_per_tile_base.data.sort_values(
        "tile_id"
    ).reset_index()
    visits_per_tile_alternative_sorted = visits_per_tile_alternative.data.sort_values(
        "tile_id"
    ).reset_index()
    relative_base = (
        visits_per_tile_base_sorted["visits"]
        / visits_per_tile_base_sorted["visits"].sum()
    )
    relative_alternative = (
        visits_per_tile_alternative_sorted["visits"]
        / visits_per_tile_alternative_sorted["visits"].sum()
    )
    deviation_from_base = pd.DataFrame(
        {
            const.TILE_ID: visits_per_tile_base_sorted[const.TILE_ID],
            "relative_base": relative_base,
            "relative_alternative": relative_alternative,
        }
    )

    deviation_from_base["deviation"] = deviation_from_base.apply(
        lambda x: symmetric_perc_error(
            x["relative_alternative"], x["relative_base"], keep_direction=True
        ),
        axis=1,
    )

    # merge count and tessellation
    counts_per_tile_gdf = pd.merge(
        tessellation,
        deviation_from_base,
        how="left",
        left_on=const.TILE_ID,
        right_on=const.TILE_ID,
    )

    map, legend = plot.choropleth_map(
        counts_per_tile_gdf,
        "deviation",
        scale_title="deviation of relative counts per tile from base",
        aliases=["Tile ID", "Tile Name", "deviation"],
        is_cmap_diverging=True,
        min_scale=-2,
        max_scale=2,
        layer_name="Deviation",
    )
    min_scale = min(
        counts_per_tile_gdf["relative_base"].min(),
        counts_per_tile_gdf["relative_alternative"].min(),
    )
    max_scale = max(
        counts_per_tile_gdf["relative_base"].max(),
        counts_per_tile_gdf["relative_alternative"].max(),
    )

    map, legend_base = plot.choropleth_map(
        counts_per_tile_gdf,
        "relative_base",
        scale_title="relative counts per tile",
        aliases=["Tile ID", "Tile Name", "relative counts"],
        is_cmap_diverging=False,
        map=map,
        layer_name="Relative visits base",
        show=False,
        min_scale=min_scale,
        max_scale=max_scale,
        cmap=const.BASE_CMAP,
    )
    map, legend_alternative = plot.choropleth_map(
        counts_per_tile_gdf,
        "relative_alternative",
        scale_title="relative counts per tile",
        aliases=["Tile ID", "Tile Name", "relative counts"],
        is_cmap_diverging=False,
        map=map,
        layer_name="Relative visits alternative",
        show=False,
        min_scale=min_scale,
        max_scale=max_scale,
        cmap=const.ALT_CMAP,
    )

    folium.LayerControl(collapsed=False).add_to(map)

    map.save(os.path.join(temp_map_folder, "visits_per_tile_map.html"))

    legend_html_deviation = v_utils.fig_to_html(legend)
    legend_html_base = v_utils.fig_to_html(legend_base)
    legend_html_alternative = v_utils.fig_to_html(legend_alternative)
    plt.close()
    return [
        v_utils.resize_width(legend_html_deviation, 100),
        v_utils.resize_width(legend_html_base, 100),
        v_utils.resize_width(legend_html_alternative, 100),
    ]


def render_visits_per_tile_cumsum(
    counts_per_tile: DfSection,
    counts_per_tile_alternative: Optional[DfSection] = None,
    diagonal: bool = False,
) -> str:
    df_cumsum = counts_per_tile.cumsum
    if counts_per_tile_alternative:
        df_cumsum_alternative = counts_per_tile_alternative.cumsum
    else:
        df_cumsum_alternative = None

    chart = plot.linechart_new(
        data=df_cumsum,
        x="n",
        y="cum_perc",
        data_alternative=df_cumsum_alternative,
        x_axis_label="Number of tiles",
        y_axis_label="Cumulated sum of visits per tile",
        add_diagonal=diagonal,
    )
    html = v_utils.fig_to_html(chart)
    plt.close()
    return html


def render_most_freq_tiles_ranking(visits_per_tile: DfSection, top_x: int = 10) -> str:
    topx_tiles = visits_per_tile.data.nlargest(top_x, "visits")
    topx_tiles["rank"] = list(range(1, len(topx_tiles) + 1))
    labels = (
        topx_tiles["rank"].astype(str)
        + ": "
        + topx_tiles[const.TILE_NAME]
        + "(Id: "
        + topx_tiles[const.TILE_ID]
        + ")"
    )

    ranking = plot.ranking(
        round(topx_tiles.visits),
        "number of visits per tile",
        y_labels=labels,
        margin_of_error=visits_per_tile.margin_of_error_laplace,
    )
    html_ranking = v_utils.fig_to_html(ranking)
    plt.close(ranking)
    return html_ranking


def render_most_freq_tiles_ranking_benchmark(
    visits_per_tile_base: DfSection,
    visits_per_tile_alternative: DfSection,
    top_x: int = 10,
) -> str:
    topx_tiles_base = visits_per_tile_base.data.nlargest(top_x, "visits")
    topx_tiles_base["visits"] = (
        topx_tiles_base["visits"] / topx_tiles_base["visits"].sum() * 100
    )
    topx_tiles_alternative = visits_per_tile_alternative.data.nlargest(top_x, "visits")
    topx_tiles_alternative["visits"] = (
        topx_tiles_alternative["visits"] / topx_tiles_alternative["visits"].sum() * 100
    )
    topx_tiles_merged = topx_tiles_base.merge(
        topx_tiles_alternative,
        on=["tile_id", "tile_name"],
        how="outer",
        suffixes=("_base", "_alternative"),
    )
    topx_tiles_merged.sort_values(by=["visits_base", "visits_alternative"])
    topx_tiles_merged["rank"] = list(range(1, len(topx_tiles_merged) + 1))
    labels = (
        topx_tiles_merged["rank"].astype(str)
        + ": "
        + topx_tiles_merged[const.TILE_NAME]
        + "(Id: "
        + topx_tiles_merged[const.TILE_ID]
        + ")"
    )

    ranking = plot.ranking(
        x=topx_tiles_merged.visits_base,
        x_axis_label="percentage of visits per tile",
        y_labels=labels,
        x_alternative=topx_tiles_merged.visits_alternative,
        margin_of_error=visits_per_tile_base.margin_of_error_laplace,
        margin_of_error_alternative=visits_per_tile_alternative.margin_of_error_laplace,
    )
    html_ranking = v_utils.fig_to_html(ranking)
    plt.close(ranking)
    return html_ranking


def render_visits_per_time_tile(
    counts_per_tile_timewindow: DfSection, tessellation: GeoDataFrame, threshold: float
) -> str:
    data = counts_per_tile_timewindow.data
    moe_perc_per_tile_timewindow = (
        counts_per_tile_timewindow.margin_of_error_laplace / data
    )

    if data is None:
        return None

    data[moe_perc_per_tile_timewindow > threshold] = None

    output_html = ""
    if "weekday" in data.columns:
        output_html += "<h4>Weekday</h4>"
        output_html += _create_timewindow_segment(data.loc[:, "weekday"], tessellation)

    if "weekend" in data.columns:
        output_html += "<h4>Weekend</h4>"
        output_html += _create_timewindow_segment(data.loc[:, "weekend"], tessellation)
    plt.close()
    return output_html


def render_visits_per_time_tile_benchmark(
    counts_per_tile_timewindow: DfSection,
    counts_per_tile_timewindow_alternative: DfSection,
    tessellation: GeoDataFrame,
) -> str:
    data = counts_per_tile_timewindow.data
    data_alternative = counts_per_tile_timewindow_alternative.data

    if data is None:
        return None
    if data_alternative is None:
        return None

    output_html = ""
    if "weekday" in data.columns:
        weekday_base = data.loc[:, "weekday"] / data.loc[:, "weekday"].sum().sum()
        weekday_alternative = (
            data_alternative.loc[:, "weekday"]
            / data_alternative.loc[:, "weekday"].sum().sum()
        )

        deviation_values = [
            [
                symmetric_perc_error(alt, base, keep_direction=True)
                for alt, base in zip(alt_array, base_array)
            ]
            for alt_array, base_array in zip(
                weekday_alternative.values, weekday_base.values
            )
        ]
        deviation = pd.DataFrame(deviation_values)
        deviation.set_index(weekday_base.index, inplace=True)
        deviation.columns = weekday_base.columns

        output_html += "<h4>Weekday</h4>"
        output_html += _create_timewindow_segment_benchmark(deviation, tessellation)

    if "weekend" in data.columns:
        weekend_base = data.loc[:, "weekend"] / data.loc[:, "weekend"].sum().sum()
        weekend_alternative = (
            data_alternative.loc[:, "weekend"]
            / data_alternative.loc[:, "weekend"].sum().sum()
        )
        deviation = [
            [
                symmetric_perc_error(alt, base, keep_direction=True)
                for alt, base in zip(alt_array, base_array)
            ]
            for alt_array, base_array in zip(
                weekend_alternative.values, weekend_base.values
            )
        ]
        deviation = pd.DataFrame(deviation)
        deviation.index = weekend_base.index
        deviation.columns = weekend_base.columns

        output_html += "<h4>Weekend</h4>"
        output_html += _create_timewindow_segment_benchmark(deviation, tessellation)
    plt.close()
    return output_html


def _create_timewindow_segment(df: pd.DataFrame, tessellation: GeoDataFrame) -> str:
    visits_choropleth = plot.multi_choropleth_map(df, tessellation)

    tile_means = df.mean(axis=1)
    dev_from_avg = df.div(tile_means, axis=0)
    deviation_choropleth = plot.multi_choropleth_map(
        dev_from_avg, tessellation, is_cmap_diverging=True, min_scale=0, vcenter=1
    )
    html = f"""<h4>Number of visits</h4>
        {v_utils.fig_to_html_as_png(visits_choropleth)}
        <h4>Deviation from tile average</h4>
        <div><p>The average of each cell 
        over all time windows equals 1 (100% of average traffic). 
        A value of < 1 (> 1) means that a tile is visited less (more) frequently in this time window than it is on average.</p></div>
        {v_utils.fig_to_html_as_png(deviation_choropleth)}"""  # svg might get too large
    plt.close(visits_choropleth)
    plt.close(deviation_choropleth)
    return html


def _create_timewindow_segment_benchmark(
    df_deviation: pd.DataFrame, tessellation: GeoDataFrame
) -> str:
    visits_choropleth = plot.multi_choropleth_map(
        df_deviation, tessellation, is_cmap_diverging=True, min_scale=-2, max_scale=2
    )
    html = f"""<h4>Deviation from base</h4>
        {v_utils.fig_to_html_as_png(visits_choropleth)}"""  # svg might get too large
    plt.close(visits_choropleth)
    return html



from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

from dp_mobility_report import constants as const
from dp_mobility_report.model.section import DfSection, DictSection, SeriesSection
from dp_mobility_report.report.html.html_utils import (
    all_available_measures,
    fmt,
    fmt_moe,
    get_template,
    render_benchmark_summary,
    render_eps,
    render_moe_info,
    render_summary,
)

if TYPE_CHECKING:
    from dp_mobility_report import DpMobilityReport

from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

from dp_mobility_report.visualization import plot, v_utils

if TYPE_CHECKING:
    from dp_mobility_report import BenchmarkReport


def render_overview(dpmreport: "DpMobilityReport") -> str:
    args: dict = {}
    report = dpmreport.report

    if const.DS_STATISTICS not in dpmreport.analysis_exclusion:
        args["dataset_stats_table"] = render_dataset_statistics(
            report[const.DS_STATISTICS]
        )
        args["dataset_stats_eps"] = (
            render_eps(report[const.DS_STATISTICS].privacy_budget),
        )

    if const.MISSING_VALUES not in dpmreport.analysis_exclusion:
        args["missing_values_table"] = render_missing_values(
            report[const.MISSING_VALUES]
        )
        args["missing_values_eps"] = (
            render_eps(report[const.MISSING_VALUES].privacy_budget),
        )

    if const.TRIPS_OVER_TIME not in dpmreport.analysis_exclusion:
        args["trips_over_time_eps"] = render_eps(
            report[const.TRIPS_OVER_TIME].privacy_budget
        )
        args["trips_over_time_moe"] = fmt_moe(
            report[const.TRIPS_OVER_TIME].margin_of_error_laplace
        )
        args["trips_over_time_info"] = render_trips_over_time_info(
            report[const.TRIPS_OVER_TIME].datetime_precision
        )
        args["trips_over_time_linechart"] = render_trips_over_time(
            report[const.TRIPS_OVER_TIME]
        )
        args["trips_over_time_moe_info"] = render_moe_info(
            report[const.TRIPS_OVER_TIME].margin_of_error_expmech
        )
        args["trips_over_time_summary_table"] = render_summary(
            report[const.TRIPS_OVER_TIME].quartiles
        )

    if const.TRIPS_PER_WEEKDAY not in dpmreport.analysis_exclusion:
        args["trips_per_weekday_eps"] = render_eps(
            report[const.TRIPS_PER_WEEKDAY].privacy_budget
        )
        args["trips_per_weekday_moe"] = fmt_moe(
            report[const.TRIPS_PER_HOUR].margin_of_error_laplace
        )

        args["trips_per_weekday_barchart"] = render_trips_per_weekday(
            report[const.TRIPS_PER_WEEKDAY]
        )

    if const.TRIPS_PER_HOUR not in dpmreport.analysis_exclusion:
        args["trips_per_hour_eps"] = render_eps(
            report[const.TRIPS_PER_HOUR].privacy_budget
        )
        args["trips_per_hour_moe"] = fmt_moe(
            report[const.TRIPS_PER_HOUR].margin_of_error_laplace
        )

        args["trips_per_hour_linechart"] = render_trips_per_hour(
            report[const.TRIPS_PER_HOUR].data, margin_of_error=None
        )

    template_structure = get_template("overview_segment.html")
    return template_structure.render(args)


def render_benchmark_overview(benchmark: "BenchmarkReport") -> str:

    args: dict = {}
    report_base = benchmark.report_base.report
    report_alternative = benchmark.report_alternative.report
    template_measures = get_template("similarity_measures.html")

    if const.DS_STATISTICS not in benchmark.analysis_exclusion:
        args["dataset_stats_table"] = render_benchmark_dataset_statistics(
            report_base[const.DS_STATISTICS],
            report_alternative[const.DS_STATISTICS],
            benchmark.smape,
        )
        args["dataset_stats_eps"] = (
            render_eps(report_base[const.DS_STATISTICS].privacy_budget),
            render_eps(report_alternative[const.DS_STATISTICS].privacy_budget),
        )

    if const.MISSING_VALUES not in benchmark.analysis_exclusion:
        args["missing_values_table"] = render_benchmark_missing_values(
            report_base[const.MISSING_VALUES],
            report_alternative[const.MISSING_VALUES],
            benchmark.smape,
        )
        args["missing_values_eps"] = (
            render_eps(report_base[const.MISSING_VALUES].privacy_budget),
            render_eps(report_alternative[const.MISSING_VALUES].privacy_budget),
        )

    if const.TRIPS_OVER_TIME not in benchmark.analysis_exclusion:
        args["trips_over_time_eps"] = (
            render_eps(report_base[const.TRIPS_OVER_TIME].privacy_budget),
            render_eps(report_alternative[const.TRIPS_OVER_TIME].privacy_budget),
        )
        args["trips_over_time_moe"] = (
            fmt_moe(report_base[const.TRIPS_OVER_TIME].margin_of_error_laplace),
            fmt_moe(report_alternative[const.TRIPS_OVER_TIME].margin_of_error_laplace),
        )
        args["trips_over_time_info"] = render_trips_over_time_info(
            report_base[const.TRIPS_OVER_TIME].datetime_precision
        )
        args["trips_over_time_linechart"] = render_benchmark_trips_over_time(
            report_base[const.TRIPS_OVER_TIME],
            report_alternative[const.TRIPS_OVER_TIME],
        )
        args["trips_over_time_moe_info"] = render_moe_info(
            report_base[const.TRIPS_OVER_TIME].margin_of_error_expmech
        )
        args["trips_over_time_summary_table"] = render_benchmark_summary(
            report_base[const.TRIPS_OVER_TIME].quartiles,
            report_alternative[const.TRIPS_OVER_TIME].quartiles,
        )
        args["trips_over_time_measure"] = template_measures.render(
            all_available_measures(const.TRIPS_OVER_TIME, benchmark)
        )

    if const.TRIPS_PER_WEEKDAY not in benchmark.analysis_exclusion:
        args["trips_per_weekday_eps"] = (
            render_eps(report_base[const.TRIPS_PER_WEEKDAY].privacy_budget),
            render_eps(report_alternative[const.TRIPS_PER_WEEKDAY].privacy_budget),
        )
        args["trips_per_weekday_moe"] = (
            fmt_moe(report_base[const.TRIPS_PER_WEEKDAY].margin_of_error_laplace),
            fmt_moe(
                report_alternative[const.TRIPS_PER_WEEKDAY].margin_of_error_laplace
            ),
        )

        args["trips_per_weekday_barchart"] = render_benchmark_trips_per_weekday(
            report_base[const.TRIPS_PER_WEEKDAY],
            report_alternative[const.TRIPS_PER_WEEKDAY],
        )
        args["trips_per_weekday_measure"] = template_measures.render(
            all_available_measures(const.TRIPS_PER_WEEKDAY, benchmark)
        )

    if const.TRIPS_PER_HOUR not in benchmark.analysis_exclusion:
        args["trips_per_hour_eps"] = (
            render_eps(report_base[const.TRIPS_PER_HOUR].privacy_budget),
            render_eps(report_alternative[const.TRIPS_PER_HOUR].privacy_budget),
        )
        args["trips_per_hour_moe"] = (
            fmt_moe(report_base[const.TRIPS_PER_HOUR].margin_of_error_laplace),
            fmt_moe(report_alternative[const.TRIPS_PER_HOUR].margin_of_error_laplace),
        )

        dataset = np.append(
            np.repeat("base", len(report_base[const.TRIPS_PER_HOUR].data)),
            np.repeat(
                "alternative", len(report_alternative[const.TRIPS_PER_HOUR].data)
            ),
        )
        combined_trips_per_hour = pd.concat(
            [
                report_base[const.TRIPS_PER_HOUR].data,
                report_alternative[const.TRIPS_PER_HOUR].data,
            ]
        )
        combined_trips_per_hour["dataset"] = dataset
        args["trips_per_hour_linechart"] = render_trips_per_hour(
            combined_trips_per_hour, margin_of_error=None, style="dataset"
        )
        args["trips_per_hour_measure"] = template_measures.render(
            all_available_measures(const.TRIPS_PER_HOUR, benchmark)
        )

    template_structure = get_template("overview_segment_benchmark.html")
    return template_structure.render(args)


def render_dataset_statistics(dataset_statistics: DictSection) -> str:
    moe = dataset_statistics.margin_of_errors_laplace
    data = dataset_statistics.data
    dataset_stats_list = [
        {
            "name": "Number of records",
            "estimate": fmt(data[const.N_RECORDS], target_type=int),
            "margin_of_error": fmt_moe(moe[const.N_RECORDS]),
        },
        {
            "name": "Distinct trips",
            "estimate": fmt(data[const.N_TRIPS], target_type=int),
            "margin_of_error": fmt_moe(moe[const.N_TRIPS]),
        },
        {
            "name": "Number of complete trips (start and and point)",
            "estimate": fmt(data[const.N_COMPLETE_TRIPS], target_type=int),
            "margin_of_error": fmt_moe(moe[const.N_COMPLETE_TRIPS]),
        },
        {
            "name": "Number of incomplete trips (single point)",
            "estimate": fmt(data[const.N_INCOMPLETE_TRIPS], target_type=int),
            "margin_of_error": fmt_moe(moe[const.N_INCOMPLETE_TRIPS]),
        },
        {
            "name": "Distinct users",
            "estimate": fmt(data[const.N_USERS], target_type=int),
            "margin_of_error": fmt_moe(moe[const.N_USERS]),
        },
        {
            "name": "Distinct locations (lat & lon combination)",
            "estimate": fmt(data[const.N_LOCATIONS], target_type=int),
            "margin_of_error": fmt_moe(moe[const.N_LOCATIONS]),
        },
    ]

    # create html from template
    template_table = get_template("table_conf_interval.html")
    dataset_stats_html = template_table.render(
        rows=dataset_stats_list,
    )
    return dataset_stats_html


def render_benchmark_dataset_statistics(
    dataset_statistics_base: DictSection,
    dataset_statistics_alternative: DictSection,
    smape: dict,
) -> str:
    moe_base = dataset_statistics_base.margin_of_errors_laplace
    moe_alternative = dataset_statistics_alternative.margin_of_errors_laplace
    data_base = dataset_statistics_base.data
    data_alternative = dataset_statistics_alternative.data

    dataset_stats_list = [
        {
            "name": "Number of records",
            "estimate": (
                fmt(data_base[const.N_RECORDS], target_type=int),
                fmt(data_alternative[const.N_RECORDS], target_type=int),
            ),
            "margin_of_error": (
                fmt_moe(moe_base[const.N_RECORDS]),
                fmt_moe(moe_alternative[const.N_RECORDS]),
            ),
            const.SMAPE: fmt(smape[const.N_RECORDS], target_type=float),
        },
        {
            "name": "Distinct trips",
            "estimate": (
                fmt(data_base[const.N_TRIPS], target_type=int),
                fmt(data_alternative[const.N_TRIPS], target_type=int),
            ),
            "margin_of_error": (
                fmt_moe(moe_base[const.N_TRIPS]),
                fmt_moe(moe_alternative[const.N_TRIPS]),
            ),
            const.SMAPE: fmt(smape[const.N_TRIPS], target_type=float),
        },
        {
            "name": "Number of complete trips (start and and point)",
            "estimate": (
                fmt(data_base[const.N_COMPLETE_TRIPS], target_type=int),
                fmt(data_alternative[const.N_COMPLETE_TRIPS], target_type=int),
            ),
            "margin_of_error": (
                fmt_moe(moe_base[const.N_COMPLETE_TRIPS]),
                fmt_moe(moe_alternative[const.N_COMPLETE_TRIPS]),
            ),
            const.SMAPE: fmt(smape[const.N_COMPLETE_TRIPS], target_type=float),
        },
        {
            "name": "Number of incomplete trips (single point)",
            "estimate": (
                fmt(data_base[const.N_INCOMPLETE_TRIPS], target_type=int),
                fmt(data_alternative[const.N_INCOMPLETE_TRIPS], target_type=int),
            ),
            "margin_of_error": (
                fmt_moe(moe_base[const.N_INCOMPLETE_TRIPS]),
                fmt_moe(moe_alternative[const.N_INCOMPLETE_TRIPS]),
            ),
            const.SMAPE: fmt(smape[const.N_INCOMPLETE_TRIPS], target_type=float),
        },
        {
            "name": "Distinct users",
            "estimate": (
                fmt(data_base[const.N_USERS], target_type=int),
                fmt(data_alternative[const.N_USERS], target_type=int),
            ),
            "margin_of_error": (
                fmt_moe(moe_base[const.N_USERS]),
                fmt_moe(moe_alternative[const.N_USERS]),
            ),
            const.SMAPE: fmt(smape[const.N_USERS], target_type=float),
        },
        {
            "name": "Distinct locations (lat & lon combination)",
            "estimate": (
                fmt(data_base[const.N_LOCATIONS], target_type=int),
                fmt(data_alternative[const.N_LOCATIONS], target_type=int),
            ),
            "margin_of_error": (
                fmt_moe(moe_base[const.N_LOCATIONS]),
                fmt_moe(moe_alternative[const.N_LOCATIONS]),
            ),
            const.SMAPE: fmt(smape[const.N_LOCATIONS], target_type=float),
        },
    ]

    # create html from template
    template_table = get_template("table_conf_interval_benchmark.html")
    dataset_stats_html = template_table.render(
        rows=dataset_stats_list,
    )
    return dataset_stats_html


def render_missing_values(missing_values: DictSection) -> str:
    moe = round(missing_values.margin_of_error_laplace, 1)
    data = missing_values.data
    missing_values_list = [
        {
            "name": "User ID (uid)",
            "estimate": fmt(data[const.UID], target_type=int),
            "margin_of_error": fmt_moe(moe),
        },
        {
            "name": "Trip ID (tid)",
            "estimate": fmt(data[const.TID], target_type=int),
            "margin_of_error": fmt_moe(moe),
        },
        {
            "name": "Timestamp (datetime)",
            "estimate": fmt(data[const.DATETIME], target_type=int),
            "margin_of_error": fmt_moe(moe),
        },
        {
            "name": "Latitude (lat)",
            "estimate": fmt(data[const.LAT], target_type=int),
            "margin_of_error": fmt_moe(moe),
        },
        {
            "name": "Longitude (lng)",
            "estimate": fmt(data[const.LNG], target_type=int),
            "margin_of_error": fmt_moe(moe),
        },
    ]

    template_table = get_template("table_conf_interval.html")
    missing_values_html = template_table.render(
        privacy_budget=render_eps(missing_values.privacy_budget),
        rows=missing_values_list,
    )
    return missing_values_html


def render_benchmark_missing_values(
    missing_values_base: DictSection,
    missing_values_alternative: DictSection,
    smape: dict,
) -> str:
    moe_base = round(missing_values_base.margin_of_error_laplace, 1)
    moe_alternative = round(missing_values_alternative.margin_of_error_laplace, 1)
    data_base = missing_values_base.data
    data_alternative = missing_values_alternative.data

    missing_values_list = [
        {
            "name": "User ID (uid)",
            "estimate": (
                fmt(data_base[const.UID], target_type=int),
                fmt(data_alternative[const.UID], target_type=int),
            ),
            "margin_of_error": (fmt_moe(moe_base), fmt_moe(moe_alternative)),
            const.SMAPE: fmt(smape[const.UID], target_type=float),
        },
        {
            "name": "Trip ID (tid)",
            "estimate": (
                fmt(data_base[const.TID], target_type=int),
                fmt(data_alternative[const.TID], target_type=int),
            ),
            "margin_of_error": (fmt_moe(moe_base), fmt_moe(moe_alternative)),
            const.SMAPE: fmt(smape[const.TID], target_type=float),
        },
        {
            "name": "Timestamp (datetime)",
            "estimate": (
                fmt(data_base[const.DATETIME], target_type=int),
                fmt(data_alternative[const.DATETIME], target_type=int),
            ),
            "margin_of_error": (fmt_moe(moe_base), fmt_moe(moe_alternative)),
            const.SMAPE: fmt(smape[const.DATETIME], target_type=float),
        },
        {
            "name": "Latitude (lat)",
            "estimate": (
                fmt(data_base[const.LAT], target_type=int),
                fmt(data_alternative[const.LAT], target_type=int),
            ),
            "margin_of_error": (fmt_moe(moe_base), fmt_moe(moe_alternative)),
            const.SMAPE: fmt(smape[const.LAT], target_type=float),
        },
        {
            "name": "Longitude (lng)",
            "estimate": (
                fmt(data_base[const.LNG], target_type=int),
                fmt(data_alternative[const.LNG], target_type=int),
            ),
            "margin_of_error": (fmt_moe(moe_base), fmt_moe(moe_alternative)),
            const.SMAPE: fmt(smape[const.LNG], target_type=float),
        },
    ]

    template_table = get_template("table_conf_interval_benchmark.html")
    missing_values_html = template_table.render(
        rows=missing_values_list,
    )

    return missing_values_html


def render_trips_over_time_info(datetime_precision: str) -> str:
    return f"Timestamps have been aggregated by {datetime_precision}."


def render_trips_over_time(trips_over_time: DfSection) -> str:
    if len(trips_over_time.data) <= 14:
        chart = plot.barchart(
            x=trips_over_time.data[const.DATETIME].to_numpy(),
            y=trips_over_time.data["trips"].to_numpy(),
            margin_of_error=fmt_moe(trips_over_time.margin_of_error_laplace),
            x_axis_label="Date",
            y_axis_label="% of trips",
            rotate_label=True,
        )
        html = v_utils.fig_to_html(chart)
    else:
        chart = plot.linechart(
            data=trips_over_time.data,
            x=const.DATETIME,
            y="trips",
            x_axis_label="Date",
            y_axis_label="% of trips",
            margin_of_error=trips_over_time.margin_of_error_laplace,
            rotate_label=True,
        )
        html = v_utils.fig_to_html(chart)
    plt.close(chart)
    return html


def render_benchmark_trips_over_time(
    trips_over_time: DfSection, trips_over_time_alternative: DfSection
) -> str:
    if len(trips_over_time.data) <= 14:
        chart = plot.barchart(
            x=trips_over_time.data[const.DATETIME].to_numpy(),
            y=trips_over_time.data["trips"].to_numpy(),
            y_alternative=trips_over_time_alternative.data["trips"].to_numpy(),
            margin_of_error=fmt_moe(trips_over_time.margin_of_error_laplace),
            margin_of_error_alternative=fmt_moe(
                trips_over_time_alternative.margin_of_error_laplace
            ),
            x_axis_label="Date",
            y_axis_label="% of trips",
            rotate_label=True,
        )
        html = v_utils.fig_to_html(chart)

    else:
        chart = plot.linechart_new(
            data=trips_over_time.data,
            x=const.DATETIME,
            data_alternative=trips_over_time_alternative.data,
            y="trips",
            x_axis_label="Date",
            y_axis_label="% of trips",
            margin_of_error=trips_over_time.margin_of_error_laplace,
            margin_of_error_alternative=trips_over_time_alternative.margin_of_error_laplace,
            rotate_label=True,
        )
        html = v_utils.fig_to_html(chart)
    plt.close(chart)
    return html


def render_trips_per_weekday(trips_per_weekday: SeriesSection) -> str:
    chart = plot.barchart(
        x=trips_per_weekday.data.index.to_numpy(),
        y=trips_per_weekday.data.values,
        margin_of_error=trips_per_weekday.margin_of_error_laplace,
        x_axis_label="Weekday",
        y_axis_label="% of trips",
        rotate_label=True,
    )
    plt.close(chart)
    return v_utils.fig_to_html(chart)


def render_benchmark_trips_per_weekday(
    trips_per_weekday: SeriesSection, trips_per_weekday_alternative: SeriesSection
) -> str:
    chart = plot.barchart(
        x=trips_per_weekday.data.index.to_numpy(),
        y=trips_per_weekday.data.values,
        y_alternative=trips_per_weekday_alternative.data.values,
        margin_of_error=trips_per_weekday.margin_of_error_laplace,
        margin_of_error_alternative=trips_per_weekday_alternative.margin_of_error_laplace,
        x_axis_label="Weekday",
        y_axis_label="% of trips",
        rotate_label=True,
    )
    plt.close(chart)
    return v_utils.fig_to_html(chart)


def render_trips_per_hour(
    trips_per_hour: pd.DataFrame,
    margin_of_error: Optional[float],
    style: Optional[str] = None,
) -> str:
    chart = plot.multi_linechart(
        data=trips_per_hour,
        x=const.HOUR,
        y="perc",
        style=style,
        color=const.TIME_CATEGORY,
        x_axis_label="Hour of day",
        y_axis_label="% of trips",
        hue_order=["weekday start", "weekday end", "weekend start", "weekend end"],
        margin_of_error=margin_of_error,
    )
    html = v_utils.fig_to_html(chart)
    plt.close(chart)
    return html



import logging
import warnings
from typing import Any, List, Optional, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from pandas import DataFrame
from shapely.geometry import Point

from dp_mobility_report import constants as const


def validate_inclusion_exclusion(
    analysis_selection: Optional[List[str]], analysis_exclusion: Optional[List[str]]
) -> Tuple[Optional[List[str]], Optional[List[str]]]:

    if analysis_selection is not None and analysis_exclusion is not None:
        warnings.warn(
            "The parameter `analysis_exclusion' will be ignored, as `analysis_selection' is set as well."
        )
        analysis_exclusion = None

    if analysis_selection == [const.ALL]:
        warnings.warn(
            "['all'] as input is deprecated. Use 'None' (default) instead to include all analyses."
        )
        analysis_selection = None

    return analysis_selection, analysis_exclusion


def validate_input(
    df: DataFrame,
    tessellation: Optional[GeoDataFrame],
    privacy_budget: Optional[Union[int, float]],
    max_trips_per_user: Optional[int],
    analysis_selection: Optional[List[str]],
    analysis_exclusion: Optional[List[str]],
    budget_split: dict,
    disable_progress_bar: bool,
    evalu: bool,
    user_privacy: bool,
    timewindows: Union[List[int], np.ndarray],
    max_travel_time: Optional[int],
    bin_range_travel_time: Optional[int],
    max_jump_length: Optional[Union[int, float]],
    bin_range_jump_length: Optional[Union[int, float]],
    max_radius_of_gyration: Optional[Union[int, float]],
    bin_range_radius_of_gyration: Optional[Union[int, float]],
    max_user_tile_count: Optional[int],
    bin_range_user_tile_count: Optional[int],
    max_user_time_delta: Optional[Union[int, float]],
    bin_range_user_time_delta: Optional[Union[int, float]],
    seed_sampling: Optional[int],
) -> None:
    if not isinstance(df, DataFrame):
        raise TypeError("'df' is not a Pandas DataFrame.")

    if tessellation is None:
        warnings.warn(
            "No tessellation has been specified. All analyses based on the tessallation will be omitted."
        )

    if not ((tessellation is None) or isinstance(tessellation, GeoDataFrame)):
        raise TypeError("'tessellation' is not a Geopandas GeoDataFrame.")

    if not ((max_trips_per_user is None) or isinstance(max_trips_per_user, int)):
        raise TypeError("'max_trips_per_user' is not numeric.")
    if (max_trips_per_user is not None) and (max_trips_per_user < 1):
        raise ValueError("'max_trips_per_user' has to be greater 0.")

    if analysis_selection is not None:
        if not isinstance(analysis_selection, list):
            raise TypeError("'analysis_selection' is not a list.")

        if not set(analysis_selection).issubset(
            const.SEGMENTS_AND_ELEMENTS + [const.ALL]
        ):
            raise ValueError(
                f"Unknown analyses in {analysis_selection}. Only elements from {const.SEGMENTS_AND_ELEMENTS} are valid inputs."
            )

    if analysis_exclusion is not None:
        if not isinstance(analysis_exclusion, list):
            raise TypeError("'analysis_exclusion' is not a list.")
        if not set(analysis_exclusion).issubset(const.SEGMENTS_AND_ELEMENTS):
            raise ValueError(
                f"Unknown analyses in {analysis_exclusion}. Only elements from {const.SEGMENTS_AND_ELEMENTS} are valid inputs."
            )

    if not isinstance(budget_split, dict):
        raise TypeError("'budget_split' is not a dict.")
    if not set(budget_split.keys()).issubset(const.ELEMENTS):
        raise ValueError(
            f"Unknown analyses in {budget_split}. Only elements from {const.ELEMENTS} are valid inputs as dictionary keys."
        )
    if not all(isinstance(x, int) for x in list(budget_split.values())):
        raise ValueError(
            f"Not all elements in 'budget_split' are integers: {list(budget_split.values())}."
        )

    if not isinstance(timewindows, (list, np.ndarray)):
        raise TypeError("'timewindows' is not a list or a numpy array.")

    timewindows = (
        np.array(timewindows) if isinstance(timewindows, list) else timewindows
    )
    if not all([np.issubdtype(item, int) for item in timewindows]):
        raise TypeError("not all items of 'timewindows' are integers.")

    _validate_bool(user_privacy, f"{user_privacy=}".split("=")[0])
    _validate_bool(evalu, f"{user_privacy=}".split("=")[0])
    _validate_bool(disable_progress_bar, f"{user_privacy=}".split("=")[0])

    if privacy_budget is not None:
        _validate_numeric_greater_zero(
            privacy_budget, f"{privacy_budget=}".split("=")[0]
        )
        if (max_trips_per_user is None) & user_privacy:
            warnings.warn(
                "Input parameter `max_trips_per_user` is `None` even though a privacy budget is given. The actual maximum number of trips per user will be used according to the data, though this violates user-level Differential Privacy."
            )

    _validate_int_greater_zero(max_travel_time, f"{max_travel_time=}".split("=")[0])
    _validate_int_greater_zero(
        bin_range_travel_time, f"{bin_range_travel_time=}".split("=")[0]
    )
    _validate_numeric_greater_zero(max_jump_length, f"{max_jump_length=}".split("=")[0])
    _validate_numeric_greater_zero(
        bin_range_jump_length, f"{bin_range_jump_length=}".split("=")[0]
    )
    _validate_numeric_greater_zero(
        max_radius_of_gyration, f"{max_radius_of_gyration=}".split("=")[0]
    )
    _validate_numeric_greater_zero(
        bin_range_radius_of_gyration, f"{bin_range_radius_of_gyration=}".split("=")[0]
    )
    _validate_int_greater_zero(
        max_user_tile_count, f"{max_user_tile_count=}".split("=")[0]
    )
    _validate_int_greater_zero(
        bin_range_user_tile_count, f"{bin_range_user_tile_count=}".split("=")[0]
    )
    _validate_numeric_greater_zero(
        max_user_time_delta, f"{max_user_time_delta=}".split("=")[0]
    )
    _validate_numeric_greater_zero(
        bin_range_user_time_delta, f"{bin_range_user_time_delta=}".split("=")[0]
    )

    if not ((seed_sampling is None) or isinstance(seed_sampling, int)):
        raise TypeError("'seed_sampling' is not an integer.")
    if (seed_sampling is not None) and (seed_sampling <= 0):
        raise ValueError("'seed_sampling' has to be greater 0.")


def _validate_int_greater_zero(var: Any, name: str) -> None:
    if not ((var is None) or isinstance(var, int)):
        raise TypeError(f"{name} is not an int.")
    if (var is not None) and (var <= 0):
        raise ValueError(f"'{name}' has to be greater 0.")


def _validate_numeric_greater_zero(var: Any, name: str) -> None:
    if not ((var is None) or isinstance(var, (int, float))):
        raise TypeError(f"{name} is not numeric.")
    if (var is not None) and (var <= 0):
        raise ValueError(f"'{name}' has to be greater 0.")


def _validate_bool(var: Any, name: str) -> None:
    if not isinstance(var, bool):
        raise TypeError(f"'{name}' is not type boolean.")


# unify different input options of analyses (segments and elements) to be included / excluded as excluded elements, i.e., 'analysis_exclusion'
def clean_analysis_exclusion(
    analysis_selection: Optional[List[str]],
    analysis_exclusion: Optional[List[str]],
    has_tessellation: bool,
    has_timestamps: bool,
    has_od_flows: bool,
    has_consecutive_user_trips: bool,
) -> List[str]:
    def _remove_elements(elements: list, remove_list: list) -> list:
        return [e for e in elements if e not in remove_list]

    if analysis_selection is not None:
        analysis_exclusion = const.ELEMENTS

        # remove all elements of respective segments from analysis_exclusion that are given in analysis_selection
        if const.OVERVIEW in analysis_selection:
            analysis_exclusion = _remove_elements(
                analysis_exclusion, const.OVERVIEW_ELEMENTS
            )
        if const.PLACE_ANALYSIS in analysis_selection:
            analysis_exclusion = _remove_elements(
                analysis_exclusion, const.PLACE_ELEMENTS
            )
        if const.OD_ANALYSIS in analysis_selection:
            analysis_exclusion = _remove_elements(analysis_exclusion, const.OD_ELEMENTS)
        if const.USER_ANALYSIS in analysis_selection:
            analysis_exclusion = _remove_elements(
                analysis_exclusion, const.USER_ELEMENTS
            )

        # remove single elements
        analysis_exclusion = _remove_elements(analysis_exclusion, analysis_selection)

    elif analysis_exclusion is not None:
        # deduplicate list in case there are duplicates as input (otherwise `remove` might fail)
        analysis_exclusion = list(set(analysis_exclusion))

        if const.OVERVIEW in analysis_exclusion:
            analysis_exclusion += const.OVERVIEW_ELEMENTS
            analysis_exclusion.remove(const.OVERVIEW)

        if const.PLACE_ANALYSIS in analysis_exclusion:
            analysis_exclusion += const.PLACE_ELEMENTS
            analysis_exclusion.remove(const.PLACE_ANALYSIS)

        if const.OD_ANALYSIS in analysis_exclusion:
            analysis_exclusion += const.OD_ELEMENTS
            analysis_exclusion.remove(const.OD_ANALYSIS)

        if const.USER_ANALYSIS in analysis_exclusion:
            analysis_exclusion += const.USER_ELEMENTS
            analysis_exclusion.remove(const.USER_ANALYSIS)

    else:
        analysis_exclusion = []

    if not has_tessellation:
        # warning in validation
        analysis_exclusion += const.TESSELLATION_ELEMENTS

    if not has_timestamps:
        # warning in validation
        analysis_exclusion += const.TIMESTAMP_ANALYSES

    if not has_od_flows:
        warnings.warn(
            "There are only incomplete trips, i.e., no trips with more than a single record. OD analyses cannot be conducted, thus they are excluded from the report."
        )
        analysis_exclusion += const.OD_ELEMENTS

    if (const.USER_TIME_DELTA not in analysis_exclusion) & (
        not has_consecutive_user_trips
    ):
        warnings.warn(
            "No user has more than one trip (this is also the case, if max_trips_per_user=1). No analysis of consecutive trips of a user can be conducted and will thus be excluded."
        )
        analysis_exclusion += [const.USER_TIME_DELTA]

    # deduplicate in case analyses and segments were included
    analysis_exclusion = list(set(analysis_exclusion))

    # sort according to order
    analysis_exclusion.sort(key=lambda i: const.ELEMENTS.index(i))

    return analysis_exclusion


def clean_budget_split(budget_split: dict, analysis_exclusion: List[str]) -> dict:
    intersec = set(budget_split.keys()).intersection(analysis_exclusion)
    if len(intersec) != 0:
        warnings.warn(
            f"A `budget_split`is specified for the analyses {intersec} even though they are excluded."
            "As they will be excluded, the `budget_split` specification will be ignored for these analyses."
        )

    # remove all analyses that are excluded according to `analysis_exclusion``
    remaining_analyses = set(budget_split.keys()) - set(analysis_exclusion)
    budget_split = {analysis: budget_split[analysis] for analysis in remaining_analyses}
    return budget_split


def preprocess_tessellation(tessellation: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    if const.TILE_ID not in tessellation.columns:
        raise ValueError("Column tile_id must be present in tesesllation.")
    tessellation.loc[:, const.TILE_ID] = tessellation.tile_id.astype(str)
    if const.TILE_NAME not in tessellation.columns:
        tessellation.loc[:, const.TILE_NAME] = tessellation.tile_id
    if const.GEOMETRY not in tessellation.columns:
        raise ValueError("Column 'geometry' must be present in tessellation.")
    if type(tessellation) is not gpd.GeoDataFrame:
        try:
            tessellation = gpd.GeoDataFrame(
                tessellation, geometry="geometry", crs=const.DEFAULT_CRS
            )
        except TypeError as ex:
            raise TypeError(
                "Tessellation cannot be cast to a geopandas.GeoDataFrame."
            ) from ex
    if tessellation.crs != const.DEFAULT_CRS:
        tessellation.to_crs(const.DEFAULT_CRS, inplace=True)
    return tessellation[[const.TILE_ID, const.TILE_NAME, const.GEOMETRY]]


def _validate_columns(df: pd.DataFrame) -> pd.DataFrame:
    if const.UID not in df.columns:
        raise ValueError("Column 'uid' must be present in data.")
    df[const.UID] = df[const.UID].astype(str)
    if const.TID not in df.columns:
        raise ValueError("Column 'tid' must be present in data.")
    if const.LAT not in df.columns:
        raise ValueError("Column 'lat' must be present in data.")
    if not pd.core.dtypes.common.is_float_dtype(df[const.LAT]):
        raise TypeError("Column 'lat' is not of type float.")
    if const.LNG not in df.columns:
        raise ValueError("Column 'lng' must be present in data.")
    if not pd.core.dtypes.common.is_float_dtype(df[const.LNG]):
        raise TypeError("Column 'lng' is not of type float.")
    if const.DATETIME not in df.columns:
        raise ValueError("Column 'datetime' must be present in data.")
    if pd.core.dtypes.common.is_integer_dtype(df[const.DATETIME]):
        warnings.warn(
            "Column 'datetime' consists of type int. It is only interpreted as sequence information and thus all time related analyses are omitted."
        )
    else:
        try:
            df[const.DATETIME] = pd.to_datetime(df[const.DATETIME])
        except Exception as ex:
            raise TypeError("Column 'datetime' cannot be cast to datetime.") from ex
    return df


def preprocess_data(
    df: pd.DataFrame,
    tessellation: Optional[gpd.GeoDataFrame],
    max_trips_per_user: int,
    user_privacy: bool,
    seed: Optional[int],
) -> pd.DataFrame:
    df = _validate_columns(df)

    df.loc[:, const.ID] = range(0, len(df))

    # make sure trip ids are unique and ordered correctly
    df[const.TID] = (
        df.sort_values([const.UID, const.DATETIME])
        .groupby([const.UID, const.TID], sort=False)
        .ngroup()
    )

    # remove unnessessary columns
    columns = [const.ID, const.UID, const.TID, const.DATETIME, const.LAT, const.LNG]
    if const.TILE_ID in df.columns:
        columns.append(const.TILE_ID)
    df = df.loc[:, columns]

    # only create time realted variables if df has timestamps
    if pd.core.dtypes.common.is_datetime64_dtype(df[const.DATETIME]):
        # create time related variables
        df.loc[:, const.HOUR] = df[const.DATETIME].dt.hour
        df.loc[:, const.IS_WEEKEND] = np.select(
            [df[const.DATETIME].dt.weekday > 4], ["weekend"], default="weekday"
        )

    # remove waypoints
    df = df.sort_values(const.DATETIME).groupby(const.TID).nth([0, -1])
    df.reset_index(inplace=True)

    # assign start and end as point_type
    df[const.POINT_TYPE] = "start"
    df.sort_values(const.DATETIME, inplace=True)
    df.loc[
        df.groupby(const.TID)[const.POINT_TYPE].tail(1).index, const.POINT_TYPE
    ] = "end"

    # if tile assignment isn't already provided, recompute assignment
    if tessellation is not None:
        if const.TILE_ID not in df.columns:
            df = assign_points_to_tessellation(df, tessellation)
        else:
            logging.info(
                "'tile_id' present in data. No new assignment of points to tessellation."
            )
            df.tile_id = df.tile_id.astype(str)

    df = sample_trips(df, max_trips_per_user, user_privacy, seed)
    return df


def assign_points_to_tessellation(
    df: pd.DataFrame, tessellation: gpd.GeoDataFrame
) -> pd.DataFrame:
    gdf = gpd.GeoDataFrame(
        df,
        geometry=[Point(xy) for xy in zip(df[const.LNG], df[const.LAT])],
        crs="EPSG:4326",
    )

    # Spatial join points to polygons
    df = gpd.sjoin(
        tessellation[[const.TILE_ID, const.TILE_NAME, const.GEOMETRY]],
        gdf,
        how="right",
    )
    df.drop(["index_left", const.GEOMETRY], axis=1, inplace=True)
    return pd.DataFrame(df)


def sample_trips(
    df: pd.DataFrame,
    max_trips_per_user: int,
    user_privacy: bool,
    seed: Optional[int],
) -> pd.DataFrame:
    if user_privacy:
        if seed is not None:
            np.random.seed(seed)
        tid_sample = (
            df[[const.UID, const.TID]]
            .drop_duplicates(const.TID)
            .groupby(const.UID)[const.TID]
            .apply(
                lambda x: np.random.choice(
                    x,
                    size=(
                        max_trips_per_user if max_trips_per_user < len(x) else len(x)
                    ),
                    replace=False,
                )
            )
        )
        return df.loc[
            df[const.TID].isin(np.concatenate(tid_sample.values)),
        ]
    else:
        return df
